# audio-offset-finder
#
# Copyright (c) 2014-23 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np
import os
import types
from audio_offset_finder.cli import main, reorder_correlations
from unittest.mock import patch
from io import StringIO
import tempfile


def test_reorder_correlations():
    input_array1 = np.array([0, 1, 2, 3])
    np.testing.assert_array_equal(reorder_correlations(input_array1, 2), np.array([2, 3, 0, 1]))

    input_array2 = np.array([0, 1, 2, 3, 4])
    np.testing.assert_array_equal(reorder_correlations(input_array2, 2), np.array([2, 3, 4, 0, 1]))


def test_tool():
    temp_dir = tempfile.TemporaryDirectory()
    plot_file_path = os.path.join(temp_dir.name, "zzz.png")
    args1 = (
        "--find-offset-of tests/audio/timbl_2.mp3 --within tests/audio/timbl_1.mp3 --resolution 160 " "--trim 35 --save-plot "
    ) + plot_file_path
    with patch("sys.stdout", new=StringIO()) as fakeStdout:
        main(args1.split())
        output = fakeStdout.getvalue().strip()
        assert output, "audio_offset_finder did not produce any output"
        assert "Offset: 12.26" in output
        assert "score: 28" in output  # Different FFmpeg versions can slightly alter this value, so don't be too strict
    assert os.path.isfile(plot_file_path), "audio_offset_finder did not create a plot file"
    temp_dir.cleanup()

    args2 = "--find-offset-of tests/audio/timbl_2.mp3"
    with pytest.raises(SystemExit) as error:
        main(args2.split())
        assert error.type == SystemExit
        assert error.value.code > 0, "missing 'within' file"

    args3 = "--within tests/audio/timbl_1.mp3"
    with pytest.raises(SystemExit) as error:
        main(args3.split())
        assert error.type == SystemExit
        assert error.value.code > 0, "missing 'offset-of' file"


def test_json():
    import json

    args = "--find-offset-of tests/audio/timbl_2.mp3 --within tests/audio/timbl_1.mp3 --resolution 160 " "--trim 35 --json"
    with patch("sys.stdout", new=StringIO()) as fakeStdout:
        main(args.split())
        output = fakeStdout.getvalue().strip()
        json_array = json.loads(output)
        assert len(json_array) == 2
        assert pytest.approx(json_array["time_offset"]) == 12.26
        assert pytest.approx(json_array["standard_score"], rel=1e-2) == 28.99



# audio-offset-finder
#
# Copyright (c) 2014-23 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .audio_offset_finder import find_offset_between_files
import argparse
import sys


def main(argv):
    parser = argparse.ArgumentParser(
        description=(
            "Find the offset of one audio file within another.\n"
            "Returns the offset in seconds from the start of the 'within' file to the start of the 'offset-of' one.\n"
            "A negative offset implies that the 'within' file starts after the 'offset-of' one.\n\n"
            "Also returns the 'standard score' of the correlation peak.  Scores >10 imply a good probability that "
            "the offset is accurate to within the precision of the tool (around 0.01s, adjustable)"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument("--find-offset-of", metavar="audio file", type=str, help="Find the offset of this file...")
    parser.add_argument("--within", metavar="audio file", type=str, help="...within this file.")
    parser.add_argument("--sr", metavar="sample rate", type=int, default=8000, help="Resample to this rate before searching")
    parser.add_argument("--trim", metavar="seconds", type=int, help="Only consider the first n seconds of the audio files")
    parser.add_argument(
        "--resolution", metavar="samples", type=int, default=128, help="Resolution (maximum accuracy) of search in samples"
    )
    parser.add_argument("--show-plot", action="store_true", dest="show_plot", help="Display plot of cross-correlation results")
    parser.add_argument(
        "--save-plot",
        metavar="filename",
        dest="plot_file",
        type=str,
        help=("Save a plot of cross-correlation results to a file " "(format matches extension - png, ps, pdf, svg)"),
    )
    parser.add_argument("--json", action="store_true", dest="output_json", help="Output in JSON for further processing")
    args = parser.parse_args(argv)
    if not (args.find_offset_of and args.within):
        parser.error("Please provide input audio files")

    try:
        trim = None
        if args.trim:
            trim = int(args.trim)

        results = find_offset_between_files(
            args.within, args.find_offset_of, fs=int(args.sr), trim=trim, hop_length=int(args.resolution)
        )
    except Exception as e:
        print(e, file=sys.stderr)
        return 1

    if args.output_json:
        import json

        json_results = {"time_offset": results["time_offset"], "standard_score": results["standard_score"]}
        print(json.dumps(json_results))
    else:
        print("Offset: %s (seconds)" % str(results["time_offset"]))
        print("Standard score: %s" % str(results["standard_score"]))

    if args.show_plot or args.plot_file is not None:
        plot_results(args, results)


# Re-order the cross-correlation array so that the index of the earliest frame offset is at one end of the range
def reorder_correlations(cc, earliest_frame_offset):
    from numpy import concatenate

    num_elements = len(cc)
    return concatenate((cc[earliest_frame_offset:], cc[:earliest_frame_offset]))


def plot_results(args, results):
    import matplotlib.pyplot as pyplot
    import matplotlib.ticker as ticker

    cc = results["correlation"]
    cc_length = len(cc)
    plot_data = reorder_correlations(cc, results["earliest_frame_offset"])

    pyplot.figure(figsize=(15, 5))
    xaxis_range = range(results["earliest_frame_offset"], results["latest_frame_offset"])
    pyplot.plot(xaxis_range, plot_data)

    ax = pyplot.gca()
    # Scale x values from frame numbers to time: t = mx + c, but c=0 for a symetrical cross-correlation
    m = results["time_scale"]
    ticks_x = ticker.FuncFormatter(lambda x, pos: "{0:g}".format(x * m))
    ax.xaxis.set_major_locator(ticker.MaxNLocator(nbins="auto"))
    ax.xaxis.set_major_formatter(ticks_x)
    ax.set_xlabel("Time offset /s", fontsize="12")
    ax.set_ylabel("Cross-correlation coefficient", fontsize="12")
    ax.set(yticklabels=[])
    ax.tick_params(left=False)

    plot_title = "Offset of %s in %s" % (args.find_offset_of, args.within)
    pyplot.title(plot_title, fontsize=14)

    peak_xvalue = results["frame_offset"]
    pyplot.axvline(x=peak_xvalue, color="red", linestyle="dotted")

    if args.plot_file is not None:
        pyplot.savefig(args.plot_file)
    if args.show_plot:
        pyplot.show()


def run():
    sys.exit(main(sys.argv[1:]))


if __name__ == "__main__":
    run()



# audio-offset-finder
#
# Copyright (c) 2014-24 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from subprocess import Popen, PIPE
from scipy.io import wavfile
import librosa
import os
import tempfile
import warnings
import numpy as np


class InsufficientAudioException(Exception):
    """Thrown when there isn't enough audio to process"""

    pass


def mfcc(audio, win_length=256, nfft=512, fs=16000, hop_length=128, numcep=13):
    """Wraps the librosa MFCC routine.  Somewhat present for historical reasons at this point."""
    return [
        np.transpose(
            librosa.feature.mfcc(y=audio, sr=fs, n_fft=nfft, win_length=win_length, hop_length=hop_length, n_mfcc=numcep)
        )
    ]


def find_offset_between_files(file1, file2, fs=8000, trim=None, hop_length=128, win_length=256, nfft=512, max_frames=2000):
    """Find the offset time offset between two audio files.

    This function takes in two file paths, and (assuming they are media files with a valid audio track)
    compares them using cross-correlation of Mel Frequency Cepstral Coefficients to try and calculate
    the time offset between them, assuming that they are two recordings of the same thing.

    Parameters
    ----------
    file1: string
        A path to the reference file, in any format that FFMPEG can read
    file2: string
        A path to the comparison file, in any format that FFMPEG can read
    fs: int
        The sampling rate that the audio should be resampled to prior to MFCC calculation, in Hz
    trim: int
        The length to which input files will be truncated before processing, in seconds.  A value of "None" indicates no trimming.
    hop_length: int
        The number of samples (at the resampled rate "fs") to skip between each calculated MFCC frame
    win_length: int
        The length of the window function used to avoid transients adding spurious high frequencies to the MFCCs
    nfft: int
        The number of samples to use in the FFTs used to generate the MFCCs

    Returns
    -------
    A dict containing the following:
    time_offset (float): the most likely offset of file2 compared to file1, in seconds.  A positive value indicates that
                         file2 starts after file1
    frame_offset (int): the offset of file2 compared to file1, measured in MFCC frames
    standard_score (float): the standard score of the highest correlation coefficient in the cross-correlation curve
    correlation (numpy int array): the 1D array of correlation coefficients calculated for the two input files
    time_scale: the scalar factor that is multiplied to frame offsets to convert them to time offsets
    earliest_frame_offset (int): the earliest offset searched for a correlation.  Always negative.
    latest_frame_offset (int): the latest offset searched for a correlation.  Always positive.

    Throws
    ------
    InsufficientAudioException if the audio supplied is too short to analyse.
    """
    tmp1 = convert_and_trim(file1, fs, trim)
    tmp2 = convert_and_trim(file2, fs, trim)
    a1 = wavfile.read(tmp1, mmap=True)[1].astype(float)
    a2 = wavfile.read(tmp2, mmap=True)[1].astype(float)
    offset_dict = find_offset_between_buffers(a1, a2, fs, hop_length, win_length, nfft)
    os.remove(tmp1)
    os.remove(tmp2)
    return offset_dict


def find_offset_between_buffers(buffer1, buffer2, fs, hop_length=128, win_length=256, nfft=512, max_frames=2000):
    """Find the offset time offset between two audio files.

    This function takes in two numpy arrays (assumed to be PCM audio) and compares them using cross-correlation of
    Mel Frequency Cepstral Coefficients to try and calculate the time offset between them, assuming that they are
    two recordings of the same thing.

    Parameters
    ----------
    buffer1: numpy array
        The first audio buffer to compare
    buffer2: numpy array
        The second audio buffer to compare
    fs: int
        The sampling rate of the two audio buffers, in Hz
    hop_length: int
        The number of samples to skip between each calculated MFCC frame
    win_length: int
        The length of the window function used to avoid transients adding spurious high frequencies to the MFCCs
    nfft: int
        The number of samples to use in the FFTs used to generate the MFCCs

    Returns
    -------
    A dict containing the following:
    time_offset (float): the most likely offset of buffer2 compared to buffer1, in seconds.  A positive value indicates that
                         buffer2 starts after buffer1
    frame_offset (int): the offset of buffer2 compared to buffer1, measured in MFCC frames
    standard_score (float): the standard score of the highest correlation coefficient in the cross-correlation curve
    correlation (numpy int array): the 1D array of correlation coefficients calculated for the two input buffers
    time_scale: the scalar factor that is multiplied to frame offsets to convert them to time offsets
    earliest_frame_offset (int): the earliest offset searched for a correlation.  Always negative.
    latest_frame_offset (int): the latest offset searched for a correlation.  Always positive.

    Throws
    ------
    InsufficientAudioException if the audio supplied is too short to analyse.
    """
    mfcc1 = mfcc(buffer1, win_length=win_length, nfft=nfft, fs=fs, hop_length=hop_length, numcep=26)[0]
    mfcc2 = mfcc(buffer2, win_length=win_length, nfft=nfft, fs=fs, hop_length=hop_length, numcep=26)[0]
    mfcc1 = std_mfcc(mfcc1)
    mfcc2 = std_mfcc(mfcc2)

    # Derive correl_nframes from the length of audio supplied, to avoid buffer overruns
    correl_nframes = min(int(len(mfcc1) / 3), len(mfcc2), max_frames)
    if correl_nframes < 10:
        raise InsufficientAudioException(
            "Not enough audio to analyse - try longer clips, less trimming, or higher resolution."
        )

    c, earliest_frame_offset, latest_frame_offset = cross_correlation(mfcc1, mfcc2, nframes=correl_nframes)

    # Find the largest value in the array of cross-correlation results (the most likely offset between the buffers)
    # and then convert it into a time offset (see also the documentation for the cross_correlation() function)
    max_k_index = np.argmax(c)
    max_k_frame_offset = max_k_index
    if max_k_frame_offset > latest_frame_offset:
        max_k_frame_offset -= len(c)
    time_scale = hop_length / fs
    time_offset = (max_k_frame_offset) * time_scale

    if np.std(c) < 1e-10:
        score = np.inf
    else:
        score = (c[max_k_index] - np.mean(c)) / np.std(c)  # standard score of peak
    return {
        "time_offset": time_offset,
        "frame_offset": int(max_k_frame_offset),
        "standard_score": score,
        "correlation": c,
        "time_scale": time_scale,
        "earliest_frame_offset": int(earliest_frame_offset),
        "latest_frame_offset": int(latest_frame_offset),
    }


# returns an array in which the first half represents an offset of mfcc2 within mfcc1,
# and the second half (accessed by negative indices) vice-versa.
def cross_correlation(mfcc1, mfcc2, nframes):
    """Calculate the cross-correlation curve between two numpy arrays (assumed to be MFCCs).

    Parameters
    ----------
    mfcc1: numpy array
        The first array to correlate
    mfcc2: numpy array
        The second array to correlate
    nframes: int
        The number of frames to correlate between the two arrays

    Returns
    -------
    A numpy array containing the cross-correlation of the two arrays for each possible offset between them.
    The zeroth element contains the cross-correlation with no offset.
    The first part of the array (with negative indices) contains offsets of mfcc1 within mfcc2.
    The second part of the array (with positive indices) contains offsets of mfcc2 within mfcc1.
    This is done so that accessing the array with an index in the range o_min to o_max will return
    an appropriate cross-correlation coefficient for that offset.
    However, it means that an array index cannot simply be translated into a time offset by multiplying
    it by the hop size divided by the sample rate.  First, the distance from the start of the array
    """
    n1, mdim1 = mfcc1.shape
    n2, mdim2 = mfcc2.shape
    o_min = nframes - n2
    o_max = n1 - nframes + 1
    n = o_max - o_min
    c = np.zeros(n)
    for k in range(o_min, 0):
        cc = np.sum(np.multiply(mfcc1[:nframes], mfcc2[-k : nframes - k]), axis=0)
        c[k] = np.linalg.norm(cc)
    for k in range(0, o_max):
        cc = np.sum(np.multiply(mfcc1[k : k + nframes], mfcc2[:nframes]), axis=0)
        c[k] = np.linalg.norm(cc)
    return c, o_min, o_max


def std_mfcc(array):
    """Returns the standard score for each offset of a given numpy array"""
    return (array - np.mean(array, axis=0)) / np.std(array, axis=0)


def convert_and_trim(afile, fs, trim=None):
    """Converts the input media to a temporary 16-bit WAV file and trims it to length.

    Parameters
    ----------
    afile: string
        The input media file to process.  It must contain at least one audio track.
    fs: int
        The sample rate that the audio should be converted to during the conversion
    trim: float
        The length to which the output audio should be trimmed, in seconds.  (Audio beyond this point will be discarded.)
        A value of "None" implies no trimming.

    Returns
    -------
    A string containing the path of the processed media file.  You should delete this file after use.
    """
    tmp = tempfile.NamedTemporaryFile(mode="r+b", prefix="offset_", suffix=".wav")
    tmp_name = tmp.name
    tmp.close()

    ffmpeg_command = ["ffmpeg"]
    ffmpeg_command += ["-loglevel", "error"]
    ffmpeg_command += ["-i", afile]
    ffmpeg_command += ["-ac", "1"]
    ffmpeg_command += ["-ar", str(fs)]
    ffmpeg_command += ["-ss", "0"]
    if trim:
        ffmpeg_command += ["-t", str(trim)]
    ffmpeg_command += ["-acodec", "pcm_s16le"]
    ffmpeg_command += [tmp_name]

    psox = Popen(ffmpeg_command, stderr=PIPE, text=True)

    stdout, stderr = psox.communicate()
    if psox.returncode != 0:
        raise Exception("FFMpeg failed:\n" + stderr.strip())
    return tmp_name



# audio-offset-finder
#
# Copyright (c) 2014-24 British Broadcasting Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from audio_offset_finder.audio_offset_finder import find_offset_between_files, std_mfcc, cross_correlation
from audio_offset_finder.audio_offset_finder import InsufficientAudioException
import numpy as np
import os


def path(test_file):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), "audio", test_file))


def test_find_offset_between_files():
    # timbl_1.mp3: Full file
    # timbl_2.mp3: File truncated at 12.254 seconds
    # timbl_3.mp3: File truncated at 12.223 seconds with white noise added to it
    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_2.mp3"), hop_length=160, trim=35)
    assert results["time_offset"] == pytest.approx(12.26)
    assert results["standard_score"] == pytest.approx(28.99, rel=1e-2)

    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_3.mp3"), hop_length=160, trim=35)
    assert results["time_offset"] == pytest.approx(12.24)
    assert results["standard_score"] == pytest.approx(23.49, rel=1e-2)

    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_1.mp3"), hop_length=160, trim=35)
    assert results["time_offset"] == pytest.approx(0.0)
    assert results["standard_score"] == pytest.approx(33.79, rel=1e-2)

    results = find_offset_between_files(path("timbl_2.mp3"), path("timbl_2.mp3"), hop_length=160, trim=35)
    assert results["time_offset"] == pytest.approx(0.0)
    assert results["standard_score"] == pytest.approx(32.48, rel=1e-2)

    results = find_offset_between_files(path("timbl_2.mp3"), path("timbl_1.mp3"), hop_length=160, trim=35)
    assert results["time_offset"] == pytest.approx(-12.26)
    assert results["standard_score"] == pytest.approx(28.99, rel=1e-2)

    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_2.mp3"), hop_length=160, trim=1)
    assert results["standard_score"] == pytest.approx(2.60, rel=1e-2)  # No good results["offset"] found

    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_2.mp3"), hop_length=160)
    assert results["time_offset"] == pytest.approx(12.26)
    assert results["standard_score"] == pytest.approx(
        30.09, rel=1e-2
    )  # standard score increases with more audio in cross-correlation

    with pytest.raises(InsufficientAudioException):
        find_offset_between_files(path("timbl_1.mp3"), path("timbl_2.mp3"), hop_length=160, trim=0.1)

    results = find_offset_between_files(path("timbl_1.mp3"), path("timbl_2.mp3"), hop_length=160, max_frames=100)
    print((results["time_offset"], results["standard_score"]))

    with pytest.raises(Exception) as exception:
        find_offset_between_files(path("dummy.mp3"), path("timbl_2.mp3"), hop_length=160, trim=0.1)
    assert exception.value.args[0].startswith("FFMpeg failed:\n")
    assert exception.value.args[0].endswith("No such file or directory")

    results = find_offset_between_files(path("r4.ogg"), path("r4_excerpt.ogg"), hop_length=128, trim=20 * 60)
    assert results["time_offset"] == pytest.approx(334.608)
    assert results["standard_score"] == pytest.approx(43.37, rel=1e-2)

    results = find_offset_between_files(path("r4.ogg"), path("r4_excerpt2.ogg"), hop_length=128, trim=20 * 60)
    assert results["time_offset"] == pytest.approx(726.4)
    assert results["standard_score"] == pytest.approx(58.66, rel=1e-2)


def test_std_mfcc():
    m = np.array([[2, 3, 4], [4, 5, 5]])
    s1 = np.std([2, 4])
    s2 = np.std([3, 5])
    s3 = np.std([4, 5])
    np.testing.assert_array_equal(std_mfcc(m), np.array([[-1.0 / s1, -1.0 / s2, -0.5 / s3], [1.0 / s1, 1.0 / s2, 0.5 / s3]]))


def test_cross_correlation():
    m1 = np.array([[-0.5, -0.4, -0.4], [0.5, 0.5, 0.4], [0.1, -0.1, 0.1]])
    m2 = np.array([[0.5, 0.5, 0.4], [0.1, -0.1, 0.1], [-0.6, 0.0, -0.3]])
    c, n_min, n_max = cross_correlation(m1, m2, 2)
    assert np.argmax(c) == 1
    assert n_min == -1
    assert n_max == 2

    c, n_min, n_max = cross_correlation(m2, m1, 2)
    offset = np.argmax(c)
    if offset > len(c) / 2:  # argmax doesn't know that the cross-correlation array is centred on 0
        offset -= len(c)
    assert offset == -1
    assert n_min == -1
    assert n_max == 2

    m2 = m1
    c, n_min, n_max = cross_correlation(m1, m2, 2)
    assert np.argmax(c) == 0
    assert n_min == -1
    assert n_max == 2
